-- Very simple spell Queue ADT for Balance Druids
-- Created by Kurohoshi (Minahonda-EU)
-- v1.1.2
-- Can be used by any class, but it's designed for Balance Druids (can track weird energy from WR)
-- Designed to use in BalancePowerTracker and LibBalancePowerTracker

--[[Changelog:
v1.1.3: Changed WRname to WRspellId
v1.1.2: Improved sWR calculation for missed wraths
		Added another field to store things (missChance)
		Code improved a little
		Changed special, if 0 then 14, else 13, but it stores specialWR (0,1,2) (Merged two vars)
--]]

local version = {1,1,3};

if (SpellQueueADT and SpellQueueADT:CompareVersion(version)) then return; end;

--Initialize Global Lib
SpellQueueADT = {};

function SpellQueueADT:CompareVersion(versionTable) --Returns false if versionTable is newer than version
	for i,v in ipairs(versionTable) do
		if version[i] < v then
			return false;
		end;
	end;
	return true;
end;
function SpellQueueADT:New()
	local SpellQueue = {}
	
	local lastSpellCasting = {n=false ,num=-1,s = false,mc=1}
	local flyingQueue = {queue={},last=0}
	local specialWR = 1 --0 when this(if casting)/next(otherwise) Wr will give 14 energy
	local previousWR = -14 --energy given by the previous WR
	local WRspellId = 5176

	function SpellQueue:BeginCastingSpell(name,num,missChance)
		lastSpellCasting.n=name
		lastSpellCasting.num=num
		lastSpellCasting.mc = missChance
		lastSpellCasting.s = specialWR;
		return true
	end

	function SpellQueue:FailedCastingSpell(num)
		if num ~= lastSpellCasting.num then return false end
		lastSpellCasting.n=false
		lastSpellCasting.num=-1
		lastSpellCasting.s = false
		lastSpellCasting.mc = 1
		return true
	end
	
	function SpellQueue:FinishCastingSpell(name,num,missChance) --returns true if it was an instant spell
		if num ~= lastSpellCasting.num then --Instant
			self:AddInstantSpell(name,specialWR,missChance)
			if name == WRspellId then --13 14 energy
				specialWR  = (specialWR +1) % 3 
			end
			
			lastSpellCasting.n=false
			lastSpellCasting.num=-1
			lastSpellCasting.s = false
			lastSpellCasting.mc = 1
			return true
		else --Casted
			self:AddInstantSpell(lastSpellCasting.n,lastSpellCasting.s,lastSpellCasting.mc)
			if lastSpellCasting.n == WRspellId then
				specialWR  = (specialWR +1) % 3 
			end
						
			lastSpellCasting.n=false
			lastSpellCasting.num=-1
			lastSpellCasting.s = false
			lastSpellCasting.mc = 1
			return false
		end
	end
	
	function SpellQueue:AddInstantSpell(name,special,missChance)
		local last = flyingQueue.last + 1
		flyingQueue.last=last
		flyingQueue.queue[last]={n=name,s=special,t=GetTime(),mc=missChance};
		return true
	end
		
	function SpellQueue:RemoveFlyingSpell(name)
		local v
		for i = 1, flyingQueue.last do
			v = flyingQueue.queue[i]
			if v.n == name then
				local last = flyingQueue.last-1
				flyingQueue.last=last
				
				for j=i,last do
					flyingQueue.queue[j]=flyingQueue.queue[j+1]
				end
				
				return true
			end
		end
		return false
	end
	
	function SpellQueue:RemoveTimedOutFlyingSpell(timeFlying)
		local timeOrigin = GetTime()-timeFlying
		local v
		for i = 1, flyingQueue.last do
			v = flyingQueue.queue[i]
			if v.t >= (timeOrigin) then
				if i == 1 then return true end
				
				for j=1,flyingQueue.last do
					flyingQueue.queue[j]=flyingQueue.queue[j+i-1]
				end
				
				flyingQueue.last = flyingQueue.last-i +1				
				
				return true
			end
		end
		flyingQueue = {queue={},last=0}
		return true
	end

	function SpellQueue:Clear()
		flyingQueue = {queue={},last=0}
		lastSpellCasting.n=false
		lastSpellCasting.num=-1
		lastSpellCasting.s = false
		lastSpellCasting.mc = 1
		return true
	end

	function SpellQueue:tostring()
		local temp = ""
		
		local v
		for i = 1, flyingQueue.last do
			v = flyingQueue.queue[i]			
			temp = i.." "..v.n.." "..tostring(v.s).." \n"..temp
		end
		if lastSpellCasting.n then
			temp="Casting: "..lastSpellCasting.n.." "..tostring(lastSpellCasting.s).." \n"..temp
		else
			temp="Casting: none \n"..temp
		end
		return temp
	end
	
	function SpellQueue:First()
		if flyingQueue.last >0 then
			return flyingQueue.queue[1].n,flyingQueue.queue[1].s,flyingQueue.queue[1].mc;
		end
	end
	
	local function specialWRBeginsAt(index)
		--if energy was -14 index begins at 1
		
		local v
		for i = 1, flyingQueue.last do
			v = flyingQueue.queue[i]
			if v.n == WRspellId then
				v.s = index
				index=(index+1)%3
			end
		end
		
		specialWR = index
		
		if lastSpellCasting.n == WRspellId then
			lastSpellCasting.s=specialWR;
		end
	end
	
	local specialWRbyEnergy={
		[-13] = {[-13] = 0,[-14] = 2,[-26] = 2,[-27] = 0,},
		[-14] = {[-13] = 1,[-14] = 1,[-26] = 1,[-27] = 1,},
		[-26] = {[-13] = 1,[-14] = 0,[-26] = 0,[-27] = 1,},
		[-27] = {[-13] = 2,[-14] = 2,[-26] = 2,[-27] = 2,},
	}
	
	function SpellQueue:RecalcS(energy)
		if (energy > 0) or (energy == -15) then --If SF energy return
			return
		end 
		
		--[[if energy == -14 then --The spell just landed gave 14 energy, backtrack the other WRs
			specialWRBeginsAt(1)
		elseif energy == -26 and (previousWR== -14 or previousWR== -26) then  --(next:14) (13,13) (landed before) next will be -14
			specialWRBeginsAt(0)
		elseif energy == -26 and (previousWR == -13 or previousWR == -27) then  --(next:13) (14,13) (landed before) -- like just -14 landed 
			specialWRBeginsAt(1)
		elseif energy == -13 and (previousWR== -14 or previousWR== -26) then  --(next:13) (13) (14 or 26) next will be 13 and then 14
			specialWRBeginsAt(2)
		elseif energy == -13 and (previousWR == -13 or previousWR == -27) then  --(next:14) (13) (13 or 27) next will be 14
			specialWRBeginsAt(0)
		elseif energy == -27 then --(next:13) (13,14) (landed before) --I never got (14,13) --> next will be 13 and next 14
			specialWRBeginsAt(2)
		end--]]
		
		specialWRBeginsAt(specialWRbyEnergy[energy][previousWR])
		
		previousWR = energy
	end
	
    function SpellQueue:iterator ()
		local i,n = 0,flyingQueue.last;
		return 	function () --function next()
					i = i + 1
					if i <= n then return flyingQueue.queue[i];
					elseif i == n+1 and lastSpellCasting.n then return lastSpellCasting;
					end
				end
    end
	
	function SpellQueue:MissFlyingSpell(name)
		local v
		for i = 1, flyingQueue.last do
			v = flyingQueue.queue[i]
			if v.n == name then
				local last = flyingQueue.last-1;
				flyingQueue.last=last
				
				for j=i,last do
					flyingQueue.queue[j]=flyingQueue.queue[j+1];
				end
				
				if name == WRspellId then 
					specialWRBeginsAt(v.s)
				end
				return true
			end
		end
		return false
	end
	
	return SpellQueue
end
